part of datepicker;

/// Update the year view selected date value from details value.
/// Notify the value changes to year view painter for repaint the painter.
void _updateYearViewSelection(_IPickerYearView view, _PickerStateArgs details) {
  view.selectedDate = _getSelectedValue(view._selectionMode, details);
  view._selectionNotifier.value = !view._selectionNotifier.value;
}

/// Return the selected value from state details based on picker selection mode.
dynamic _getSelectedValue(
    DateRangePickerSelectionMode mode, _PickerStateArgs details) {
  if (mode == DateRangePickerSelectionMode.single) {
    return details._selectedDate;
  } else if (mode == DateRangePickerSelectionMode.multiple) {
    return _cloneList(details._selectedDates);
  } else if (mode == DateRangePickerSelectionMode.range) {
    return details._selectedRange;
  } else {
    return _cloneList(details._selectedRanges);
  }
}

/// Check whether the date is in between the start and end date value.
bool _isDateWithInYearRange(DateTime startDate, DateTime endDate, DateTime date,
    _IPickerYearView view) {
  if (startDate == null || endDate == null || date == null) {
    return false;
  }

  /// Check the start date as before of end date, if not then swap
  /// the start and end date values.
  if (startDate.isAfter(endDate)) {
    final DateTime temp = startDate;
    startDate = endDate;
    endDate = temp;
  }

  /// Check the date is equal or after of the start date and
  /// the date is equal or before of the end date.
  if ((view._isSameView(endDate, date) || endDate.isAfter(date)) &&
      (view._isSameView(startDate, date) || startDate.isBefore(date))) {
    return true;
  }

  return false;
}

/// Return index of the date value in dates collection.
/// Return -1 when the date does not exist in dates collection.
int _getYearCellIndex(
    List<DateTime> dates, DateTime date, _IPickerYearView view,
    {int viewStartIndex = -1, int viewEndIndex = -1}) {
  if (date == null) {
    return -1;
  }

  viewStartIndex = viewStartIndex == -1 ? 0 : viewStartIndex;
  viewEndIndex = viewEndIndex == -1 ? dates.length - 1 : viewEndIndex;
  for (int i = viewStartIndex; i <= viewEndIndex; i++) {
    final DateTime _currentDate = dates[i];
    if (view._isSameView(date, _currentDate)) {
      return i;
    }
  }

  return -1;
}

/// Return the previous view date of the date based on picker view.
/// Eg., if picker view as year view and view date as May, 2020 then
/// it return date value as Apr, 2020.
DateTime _getPreviousDate(DateRangePickerView view, DateTime viewDate) {
  if (view == DateRangePickerView.year) {
    DateTime date = DateTime(viewDate.year, viewDate.month, 1);
    date = subtractDuration(date, const Duration(days: 1));
    return DateTime(date.year, date.month, 1);
  } else if (view == DateRangePickerView.decade) {
    DateTime date = DateTime(viewDate.year, 1, 1);
    date = subtractDuration(date, const Duration(days: 1));
    return DateTime(date.year, 1, 1);
  } else if (view == DateRangePickerView.century) {
    DateTime date = DateTime((viewDate.year ~/ 10) * 10, 1, 1);
    date = subtractDuration(date, const Duration(days: 1));
    return DateTime((date.year ~/ 10) * 10, 1, 1);
  }

  return viewDate;
}

/// Return the next view date of the date based on picker view.
/// Eg., if picker view as year view and view date as May, 2020 then
/// it return date value as Jun, 2020.
DateTime _getNextDate(DateRangePickerView view, DateTime viewEndDate) {
  final DateTime date = _getLastDate(view, viewEndDate);
  return addDuration(date, const Duration(days: 1));
}

/// Return the last date of the date based on picker view.
/// Eg., if picker view as year view and view date as May 20, 2020 then
/// it return date value as May 31, 2020.
DateTime _getLastDate(DateRangePickerView view, DateTime viewEndDate) {
  if (view == DateRangePickerView.year) {
    final DateTime date = DateTime(viewEndDate.year, viewEndDate.month + 1, 1);
    return subtractDuration(date, const Duration(days: 1));
  } else if (view == DateRangePickerView.decade) {
    final DateTime date = DateTime(viewEndDate.year + 1, 1, 1);
    return subtractDuration(date, const Duration(days: 1));
  } else if (view == DateRangePickerView.century) {
    final DateTime date = DateTime(viewEndDate.year + 10, 1, 1);
    return subtractDuration(date, const Duration(days: 1));
  }

  return viewEndDate;
}

/// Return list of int value in between start and end date index value.
List<int> _getRangeIndex(
    DateTime startDate,
    DateTime endDate,
    _IPickerYearView view,
    DateRangePickerView pickerView,
    int viewStartIndex,
    int viewEndIndex) {
  int startIndex = -1;
  int endIndex = -1;
  final List<int> _selectedIndex = <int>[];

  /// Check the start date as before of end date, if not then swap
  /// the start and end date values.
  if (startDate != null && startDate.isAfter(endDate)) {
    final DateTime temp = startDate;
    startDate = endDate;
    endDate = temp;
  }

  final DateTime viewStartDate = view.visibleDates[viewStartIndex];
  final DateTime viewEndDate =
      _getLastDate(pickerView, view.visibleDates[viewEndIndex]);
  if (startDate != null) {
    /// Assign start index as -1 when the start date before of view start date.
    if (viewStartDate.isAfter(startDate) && viewStartDate.isBefore(endDate)) {
      startIndex = -1;
    } else {
      startIndex = _getYearCellIndex(view.visibleDates, startDate, view,
          viewStartIndex: viewStartIndex, viewEndIndex: viewEndIndex);
    }
  }

  if (endDate != null) {
    /// Assign end index as visible dates length when the
    /// end date after of view end date.
    if (viewEndDate.isAfter(startDate) && viewEndDate.isBefore(endDate)) {
      endIndex = viewEndIndex + 1;
    } else {
      endIndex = _getYearCellIndex(view.visibleDates, endDate, view,
          viewStartIndex: viewStartIndex, viewEndIndex: viewEndIndex);
    }
  }

  //// If some range end date as null then it end index is start index.
  if (startIndex != -1 && endIndex == -1) {
    endIndex = startIndex;
  }

  /// Check the start index as before of end index, if not then swap
  /// the start and end index values.
  if (startIndex > endIndex) {
    final int temp = startIndex;
    startIndex = endIndex;
    endIndex = temp;
  }

  /// Add the index values in between start and end index values.
  for (int i = startIndex; i <= endIndex; i++) {
    _selectedIndex.add(i);
  }

  return _selectedIndex;
}

/// Draws the year cell on canvas based on picker view and [_IPickerYearView] value
void _drawYearCells(Canvas canvas, Size size, _IPickerYearView view,
    DateRangePickerView pickerView) {
  canvas.clipRect(Rect.fromLTWH(0, 0, size.width, size.height));
  double _width = size.width;
  final double webUIPadding = view.multiViewSpacing;
  if (view.enableMultiView) {
    _width = (_width - webUIPadding) / 2;
  }

  final double _cellWidth = _width / _IPickerYearView._maxColumnCount;
  final double _cellHeight = size.height / _IPickerYearView._maxRowCount;

  final TextStyle _textStyle =
      view.cellStyle.textStyle ?? view.datePickerTheme.cellTextStyle;
  final TextStyle _leadingCellTextStyle =
      view.cellStyle.leadingDatesTextStyle ??
          view.datePickerTheme.leadingCellTextStyle;
  final TextStyle _disabledCellTextStyle =
      view.cellStyle.disabledDatesTextStyle ??
          view.datePickerTheme.disabledCellTextStyle;
  final TextStyle _selectionTextStyle =
      view.selectionTextStyle ?? view.datePickerTheme.selectionTextStyle;

  double _xPosition = 0, _yPosition;
  final DateTime today = DateTime.now();
  view._textPainter ??= TextPainter(
      textAlign: TextAlign.start,
      textDirection: TextDirection.ltr,
      maxLines: 2,
      textWidthBasis: TextWidthBasis.longestLine);

  Decoration _yearDecoration;
  BoxPainter _boxPainter;
  const double decorationPadding = 1;
  const double selectionPadding = 3;
  final double _centerYPosition = _cellHeight / 2;

  final int visibleDatesCount = view.enableMultiView
      ? view.visibleDates.length ~/ 2
      : view.visibleDates.length;
  for (int j = 0; j < (view.enableMultiView ? 2 : 1); j++) {
    final int currentViewIndex =
        view.isRtl ? _getRtlIndex(view.enableMultiView ? 2 : 1, j) : j;
    _xPosition =
        (currentViewIndex * _width) + (currentViewIndex * webUIPadding);
    _yPosition = 0;
    List<int> selectedIndex = <int>[];
    final List<List<int>> rangeIndex = <List<int>>[];

    final int _viewStartIndex = j * visibleDatesCount;
    final int _viewEndIndex = ((j + 1) * visibleDatesCount) - 1;

    /// Calculate the selected index values based on selected date property in [_IPickerYearView]
    if (view.selectedDate != null) {
      if (view._selectionMode == DateRangePickerSelectionMode.single) {
        final int index = _getYearCellIndex(
            view.visibleDates, view.selectedDate, view,
            viewStartIndex: _viewStartIndex, viewEndIndex: _viewEndIndex);
        if (index != -1) {
          selectedIndex.add(index);
        }
      } else if (view._selectionMode == DateRangePickerSelectionMode.multiple) {
        for (int i = 0; i < view.selectedDate.length; i++) {
          final int index = _getYearCellIndex(
              view.visibleDates, view.selectedDate[i], view,
              viewStartIndex: _viewStartIndex, viewEndIndex: _viewEndIndex);
          if (index != -1) {
            selectedIndex.add(index);
          }
        }
      } else if (view._selectionMode == DateRangePickerSelectionMode.range) {
        final DateTime startDate = view.selectedDate.startDate;
        final DateTime endDate =
            view.selectedDate.endDate ?? view.selectedDate.startDate;
        selectedIndex = _getRangeIndex(startDate, endDate, view, pickerView,
            _viewStartIndex, _viewEndIndex);
        rangeIndex.add(selectedIndex);
      } else if (view._selectionMode ==
          DateRangePickerSelectionMode.multiRange) {
        for (int i = 0; i < view.selectedDate.length; i++) {
          final PickerDateRange range = view.selectedDate[i];
          final DateTime startDate = range.startDate;
          final DateTime endDate = range.endDate ?? range.startDate;
          final List<int> index = _getRangeIndex(startDate, endDate, view,
              pickerView, _viewStartIndex, _viewEndIndex);
          rangeIndex.add(index);
          for (int j = 0; j < index.length; j++) {
            selectedIndex.add(index[j]);
          }
        }
      }
    }

    for (int i = 0; i < visibleDatesCount; i++) {
      int _currentIndex = i;
      if (view.isRtl) {
        final int rowIndex = i ~/ _IPickerYearView._maxColumnCount;
        _currentIndex = _getRtlIndex(_IPickerYearView._maxColumnCount,
                i % _IPickerYearView._maxColumnCount) +
            (rowIndex * _IPickerYearView._maxColumnCount);
      }

      _currentIndex += _viewStartIndex;
      if (view.enableMultiView &&
          view._isTrailingDate(_currentIndex, _viewStartIndex)) {
        _xPosition += _cellWidth;
        if (_xPosition.round() >=
            (((currentViewIndex + 1) * _width) +
                    (currentViewIndex * webUIPadding))
                .round()) {
          _xPosition =
              (currentViewIndex * _width) + (currentViewIndex * webUIPadding);
          _yPosition += _cellHeight;
        }

        continue;
      }

      final DateTime _date = view.visibleDates[_currentIndex];
      TextStyle _style = _textStyle;
      final bool _isCurrentDate = view._isSameView(_date, today);
      _yearDecoration = view.cellStyle.cellDecoration;
      if (!view._isCurrentView(_date, j)) {
        _yearDecoration = view.cellStyle.leadingDatesDecoration;
        _style = _leadingCellTextStyle;
      }

      if (_isCurrentDate) {
        _yearDecoration = view.cellStyle.todayCellDecoration;
        _style = view.cellStyle.todayTextStyle ??
            view.datePickerTheme.todayCellTextStyle;
      }

      final bool _isSelected = selectedIndex.contains(_currentIndex);
      if (_isSelected) {
        _style = _selectionTextStyle;
      }

      final bool _isEnableDate = view._isBetweenMinMaxMonth(_date);
      if (!_isEnableDate) {
        _yearDecoration = view.cellStyle.disabledDatesDecoration;
        _style = _disabledCellTextStyle;
      }

      TextSpan _yearText = TextSpan(
        text: view._getCellText(_date),
        style: _style,
      );

      view._textPainter.text = _yearText;
      view._textPainter.layout(minWidth: 0, maxWidth: _cellWidth);

      double highlightPadding =
          view._selectionRadius == null ? 10 : view._selectionRadius;
      final double _textHalfHeight = view._textPainter.height / 2;
      if (_isSelected && _isEnableDate) {
        view._todayHighlightPaint ??= Paint();
        view._todayHighlightPaint.isAntiAlias = true;
        view._todayHighlightPaint.style = PaintingStyle.fill;
        final double maximumHighlight =
            _centerYPosition - _textHalfHeight - selectionPadding;
        if (maximumHighlight < highlightPadding) {
          highlightPadding = maximumHighlight;
        }

        /// _isSelectedDate value used to notify the year cell as selected and
        /// its selection mode as single or multiple or the [PickerDateRange]
        /// holds only start date value on range and multi range selection.
        bool _isSelectedDate =
            view._selectionMode == DateRangePickerSelectionMode.single ||
                    view._selectionMode == DateRangePickerSelectionMode.multiple
                ? true
                : false;

        /// _isStartRange value used to notify the year cell as selected and
        /// the year cell as start date cell of the [PickerDateRange].
        /// its selection mode as range or multi range.
        bool _isStartRange = false;

        /// _isEndRange value used to notify the year cell as selected and
        /// the year cell as end date cell of the [PickerDateRange].
        /// its selection mode as range or multi range.
        bool _isEndRange = false;

        /// _isBetweenRange value used to notify the year cell as selected and
        /// the year cell as in between the start and  end date cell of the
        /// [PickerDateRange]. its selection mode as range or multi range.
        bool _isBetweenRange = false;
        if (view._selectionMode == DateRangePickerSelectionMode.range ||
            view._selectionMode == DateRangePickerSelectionMode.multiRange) {
          List<int> range = <int>[];
          for (int k = 0; k < rangeIndex.length; k++) {
            range = rangeIndex[k];
            if (!range.contains(_currentIndex)) {
              continue;
            }

            if (range.length == 1) {
              _isSelectedDate = true;
            } else if (range[0] == _currentIndex) {
              if (view.isRtl) {
                _isEndRange = true;
              } else {
                _isStartRange = true;
              }
            } else if (range[range.length - 1] == _currentIndex) {
              if (view.isRtl) {
                _isStartRange = true;
              } else {
                _isEndRange = true;
              }
            } else {
              _isBetweenRange = true;
            }
          }
        }

        final Rect _rect = Rect.fromLTRB(
            _xPosition +
                (_isBetweenRange || _isEndRange ? 0 : selectionPadding),
            _yPosition + _centerYPosition - highlightPadding - _textHalfHeight,
            _xPosition +
                _cellWidth -
                (_isBetweenRange || _isStartRange ? 0 : selectionPadding),
            _yPosition + _centerYPosition + highlightPadding + _textHalfHeight);
        final double cornerRadius = _isBetweenRange
            ? 0
            : (view.selectionShape == DateRangePickerSelectionShape.circle
                ? _rect.height / 2
                : 3);
        final double leftRadius =
            _isStartRange || _isSelectedDate ? cornerRadius : 0;
        final double rightRadius =
            _isEndRange || _isSelectedDate ? cornerRadius : 0;
        if (_isSelectedDate) {
          if (view._selectionMode == DateRangePickerSelectionMode.single ||
              view._selectionMode == DateRangePickerSelectionMode.multiple) {
            view._todayHighlightPaint.color =
                view.selectionColor ?? view.datePickerTheme.selectionColor;
          } else {
            view._todayHighlightPaint.color = view.startRangeSelectionColor ??
                view.datePickerTheme.startRangeSelectionColor;
          }
        } else if (_isStartRange) {
          view._todayHighlightPaint.color = view.startRangeSelectionColor ??
              view.datePickerTheme.startRangeSelectionColor;
        } else if (_isBetweenRange) {
          _yearText = TextSpan(
            text: _yearText.text,
            style: view.rangeTextStyle ??
                view.datePickerTheme.rangeSelectionTextStyle,
          );

          view._textPainter.text = _yearText;
          view._textPainter.layout(minWidth: 0, maxWidth: _cellWidth);

          view._todayHighlightPaint.color = view.rangeSelectionColor ??
              view.datePickerTheme.rangeSelectionColor;
        } else if (_isEndRange) {
          view._todayHighlightPaint.color = view.endRangeSelectionColor ??
              view.datePickerTheme.endRangeSelectionColor;
        }

        canvas.drawRRect(
            RRect.fromRectAndCorners(_rect,
                topLeft: Radius.circular(leftRadius),
                bottomLeft: Radius.circular(leftRadius),
                bottomRight: Radius.circular(rightRadius),
                topRight: Radius.circular(rightRadius)),
            view._todayHighlightPaint);
      } else if (_yearDecoration != null) {
        _boxPainter = _yearDecoration.createBoxPainter();
        _boxPainter.paint(
            canvas,
            Offset(
                _xPosition + decorationPadding, _yPosition + decorationPadding),
            ImageConfiguration(
                size: Size(_cellWidth - (2 * decorationPadding),
                    _cellHeight - (2 * decorationPadding))));
      } else if (_isCurrentDate) {
        view._todayHighlightPaint ??= Paint();
        view._todayHighlightPaint.color = view.todayHighlightColor ??
            view.datePickerTheme.todayHighlightColor;
        view._todayHighlightPaint.isAntiAlias = true;
        view._todayHighlightPaint.strokeWidth = 1.0;
        view._todayHighlightPaint.style = PaintingStyle.stroke;
        final double maximumHighlight =
            _centerYPosition - _textHalfHeight - selectionPadding;
        if (maximumHighlight < highlightPadding) {
          highlightPadding = maximumHighlight;
        }

        final Rect _rect = Rect.fromLTRB(
            _xPosition + selectionPadding,
            _yPosition + _centerYPosition - highlightPadding - _textHalfHeight,
            _xPosition + _cellWidth - selectionPadding,
            _yPosition + _centerYPosition + highlightPadding + _textHalfHeight);
        double cornerRadius = _rect.height / 2;
        if (view.selectionShape == DateRangePickerSelectionShape.rectangle) {
          cornerRadius = 3;
        }

        canvas.drawRRect(
            RRect.fromRectAndRadius(_rect, Radius.circular(cornerRadius)),
            view._todayHighlightPaint);
      }

      double xOffset =
          _xPosition + ((_cellWidth - view._textPainter.width) / 2);
      xOffset = xOffset < 0 ? 0 : xOffset;
      double yOffset =
          _yPosition + ((_cellHeight - view._textPainter.height) / 2);
      yOffset = yOffset < 0 ? 0 : yOffset;

      if (view._mouseHoverPosition != null) {
        if (!view._isBetweenMinMaxMonth(_date)) {
          view._textPainter.paint(canvas, Offset(xOffset, yOffset));
          _xPosition += _cellWidth;
          if (_xPosition.round() >=
              (((currentViewIndex + 1) * _width) +
                      (currentViewIndex * webUIPadding))
                  .round()) {
            _xPosition =
                (currentViewIndex * _width) + (currentViewIndex * webUIPadding);
            _yPosition += _cellHeight;
          }
          continue;
        }

        if (_xPosition <= view._mouseHoverPosition.dx &&
            _xPosition + _cellWidth >= view._mouseHoverPosition.dx &&
            _yPosition - 5 <= view._mouseHoverPosition.dy &&
            (_yPosition + _cellHeight) - 5 >= view._mouseHoverPosition.dy) {
          view._todayHighlightPaint = view._todayHighlightPaint ?? Paint();
          view._todayHighlightPaint.style = PaintingStyle.fill;
          view._todayHighlightPaint.strokeWidth = 2;
          view._todayHighlightPaint.color = view.selectionColor != null
              ? view.selectionColor.withOpacity(0.4)
              : view.datePickerTheme.selectionColor.withOpacity(0.4);

          if (_centerYPosition - _textHalfHeight < highlightPadding / 2) {
            highlightPadding = (_centerYPosition - _textHalfHeight / 2) - 1;
          }

          final Rect _rect = Rect.fromLTRB(
              _xPosition + selectionPadding,
              _yPosition +
                  _centerYPosition -
                  highlightPadding -
                  _textHalfHeight,
              _xPosition + _cellWidth - selectionPadding,
              _yPosition +
                  _centerYPosition +
                  highlightPadding +
                  _textHalfHeight);
          double cornerRadius = _rect.height / 2;
          if (view.selectionShape == DateRangePickerSelectionShape.rectangle) {
            cornerRadius = 3;
          }

          canvas.drawRRect(
              RRect.fromRectAndRadius(_rect, Radius.circular(cornerRadius)),
              view._todayHighlightPaint);
        }
      }

      view._textPainter.paint(canvas, Offset(xOffset, yOffset));
      _xPosition += _cellWidth;
      if (_xPosition.round() >=
          (((currentViewIndex + 1) * _width) +
                  (currentViewIndex * webUIPadding))
              .round()) {
        _xPosition =
            (currentViewIndex * _width) + (currentViewIndex * webUIPadding);
        _yPosition += _cellHeight;
      }
    }
  }
}

List<CustomPainterSemantics> _getSemanticsBuilderForYearView(
    Size size, _IPickerYearView view) {
  final List<CustomPainterSemantics> _semanticsBuilder =
      <CustomPainterSemantics>[];
  double _left, _top, _cellWidth, _cellHeight;
  Map<String, double> _leftAndTopValue;
  int count = 1;
  double _width = size.width;
  if (view.enableMultiView) {
    count = 2;
    _width = (size.width - view.multiViewSpacing) / 2;
  }

  _cellWidth = _width / 3;
  _cellHeight = size.height / 4;
  final int datesCount = view.visibleDates.length ~/ count;
  for (int j = 0; j < count; j++) {
    final int currentViewIndex = view.isRtl ? _getRtlIndex(count, j) : j;
    _left = view.isRtl ? _width - _cellWidth : 0;
    _top = 0;
    final double startXPosition = (currentViewIndex * _width) +
        (currentViewIndex * view.multiViewSpacing);
    final int _startIndex = j * datesCount;
    for (int i = 0; i < datesCount; i++) {
      final DateTime _date = view.visibleDates[_startIndex + i];
      if (view._isTrailingDate(_startIndex + i, _startIndex)) {
        _leftAndTopValue = _getTopAndLeftValues(
            view.isRtl, _left, _top, _cellWidth, _cellHeight, _width);
        _left = _leftAndTopValue['left'];
        _top = _leftAndTopValue['top'];
        continue;
      }

      if (!view._isBetweenMinMaxMonth(_date)) {
        _semanticsBuilder.add(CustomPainterSemantics(
          rect: Rect.fromLTWH(
              startXPosition + _left, _top, _cellWidth, _cellHeight),
          properties: SemanticsProperties(
            label: _getCellText(_date, view) + 'Disabled cell',
            textDirection: TextDirection.ltr,
          ),
        ));

        _leftAndTopValue = _getTopAndLeftValues(
            view.isRtl, _left, _top, _cellWidth, _cellHeight, _width);
        _left = _leftAndTopValue['left'];
        _top = _leftAndTopValue['top'];
        continue;
      }
      _semanticsBuilder.add(CustomPainterSemantics(
        rect: Rect.fromLTWH(
            startXPosition + _left, _top, _cellWidth, _cellHeight),
        properties: SemanticsProperties(
          label: _getCellText(_date, view),
          textDirection: TextDirection.ltr,
        ),
      ));
      _leftAndTopValue = _getTopAndLeftValues(
          view.isRtl, _left, _top, _cellWidth, _cellHeight, _width);
      _left = _leftAndTopValue['left'];
      _top = _leftAndTopValue['top'];
    }
  }

  return _semanticsBuilder;
}

String _getCellText(DateTime date, _IPickerYearView view) {
  if (view.runtimeType.toString() == '_PickerYearViewPainter') {
    return DateFormat('MMMM yyyy').format(date).toString();
  } else if (view.runtimeType.toString() == '_PickerDecadeViewPainter') {
    return date.year.toString();
  } else {
    return date.year.toString() + ' to ' + (date.year + 9).toString();
  }
}
